#pragma once
#include <iostream>
#include <vector>
#include "coordinate.h"

class QuadTree {
public:
    QuadTree (AABB _boundary) : boundary(_boundary), points(std::vector<XY>()) {}
    
    // Insert a point into the quadtree
    bool insert (XY p)
    {
        // Ignore this point if it is outside the bounds of this tree
        if(!boundary.containsPoint(p)){
            return false;
        }

        // If there is space in this quad tree and if it doenst have subdivisions add the object
        if(points.size() < 4 && northWest == nullptr){
            points.push_back(p);
            return true;
        }

        // otherwise , subdivide and add the point to whichever node accepts it
        if(northWest == nullptr)
        {
            subdivide();
        }

        if(northWest->insert(p)) return true;
        if(northEast->insert(p)) return true;
        if(southWest->insert(p)) return true;
        if(southEast->insert(p)) return true;


        return false;


    }
    
    std::vector<XY> queryRange(AABB& range)
    {
        // prepare a list of results
        auto  pointsInRange = std::vector<XY>();

        // abort if the range does not intersect with this quad
        if(!boundary.intersectsAABB(range))
            return pointsInRange;
        
        // check object at this level
        for (unsigned int p = 0; p < points.size(); p++ )
        {
            if(range.containsPoint(points[p]))
                pointsInRange.push_back(points[p]);
        }

        // exit if there are no children
        if(northWest == nullptr)
            return pointsInRange;

        // Otherwise, add the points from the children
        auto points = northWest->queryRange(range);
        for ( auto point : points){
            pointsInRange.push_back(point);
        }
        points = northEast->queryRange(range);
        for(auto point : points){
            pointsInRange.push_back(point);
        }
        points = southWest->queryRange(range);
        for( auto point : points){
            pointsInRange.push_back(point);
        }
        points = southEast->queryRange(range);
        for(auto point : points){
            pointsInRange.push_back(point);
        }

        return pointsInRange;

    }


    void print(int indent = 0) {
        if( indent == 0)
        std::cout << "==== Printing QuadTree ====" << std::endl;
        
        for ( int i = 0; i < indent; i++)
            std::cout << "*";
        std::cout << "* " << boundary << std::endl;
        
        // list points 
        for( auto point : points ){
            std::cout << "| " << point << std::endl;
        }

        if(northWest == nullptr)
            return;    
        
        northWest->print(indent +1);
        northEast->print(indent +1);
        southWest->print(indent +1);
        southEast->print(indent +1);
    

    }
private:
    const int NODE_CAPACITY = 4;

    // Bounding box representing the boundaries of this 
    // quad tree;
    AABB boundary ;
    
    
    // Points in this tree
    std::vector<XY> points;

    QuadTree* northWest = nullptr;
    QuadTree* northEast = nullptr;
    QuadTree* southWest = nullptr;
    QuadTree* southEast = nullptr;

    // Create four children that fully divide this quad into four quads of equal area
    void subdivide()
    {
        float Quarter_halfDimension = boundary.halfDimension / 4;
        float x = boundary.center.x;
        float y = boundary.center.y;
        northWest = new QuadTree(AABB({x - (boundary.halfDimension / 2), y + (boundary.halfDimension / 2)}, Quarter_halfDimension));
        southWest = new QuadTree(AABB({x - (boundary.halfDimension / 2), y - (boundary.halfDimension / 2) }, Quarter_halfDimension ));
        northEast = new QuadTree(AABB({x + (boundary.halfDimension / 2), y + (boundary.halfDimension / 2) }, Quarter_halfDimension ));
        southEast = new QuadTree(AABB({x + (boundary.halfDimension / 2), y - (boundary.halfDimension / 2) }, Quarter_halfDimension ));
    }
    

};